# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: frameserver.proto
"""Generated protocol buffer code."""
from __future__ import annotations

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="frameserver.proto",
    package="frameserver",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x11\x66rameserver.proto\x12\x0b\x66rameserver"I\n\x16\x46\x65tchSceneDataResponse\x12!\n\x05scene\x18\x01 \x01(\x0b\x32\x12.frameserver.Scene\x12\x0c\n\x04path\x18\x02 \x01(\t"[\n\x05Scene\x12\x0c\n\x04name\x18\x01 \x01(\t\x12*\n\nanimations\x18\x02 \x03(\x0b\x32\x16.frameserver.Animation\x12\x18\n\x10\x62\x61\x63kground_color\x18\x03 \x01(\t"\xfd\x02\n\tAnimation\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x64uration\x18\x02 \x01(\x02\x12\x17\n\x0f\x65\x61sing_function\x18\x03 \x01(\t\x12G\n\x14\x61ttribute_tween_data\x18\x04 \x03(\x0b\x32).frameserver.Animation.AttributeTweenData\x12\x43\n\x12mobject_tween_data\x18\x05 \x03(\x0b\x32\'.frameserver.Animation.MobjectTweenData\x12\x1d\n\x15\x66lickered_mobject_ids\x18\x06 \x03(\t\x1aM\n\x12\x41ttributeTweenData\x12\x11\n\tattribute\x18\x01 \x01(\t\x12\x12\n\nstart_data\x18\x02 \x03(\x02\x12\x10\n\x08\x65nd_data\x18\x03 \x03(\x02\x1a;\n\x10MobjectTweenData\x12\n\n\x02id\x18\x01 \x01(\t\x12\x1b\n\x13root_mobject_offset\x18\x02 \x03(\x02"(\n\x07Updater\x12\x1d\n\x15\x66lickered_mobject_ids\x18\x01 \x03(\t"k\n\x0c\x46rameRequest\x12\x11\n\tend_index\x18\x01 \x01(\x05\x12\x15\n\rfirst_request\x18\x02 \x01(\x08\x12\x17\n\x0f\x61nimation_index\x18\x03 \x01(\x05\x12\x18\n\x10\x61nimation_offset\x18\x04 \x01(\x02"u\n\x05Style\x12\x12\n\nfill_color\x18\x01 \x01(\t\x12\x14\n\x0c\x66ill_opacity\x18\x02 \x01(\x02\x12\x14\n\x0cstroke_color\x18\x03 \x01(\t\x12\x16\n\x0estroke_opacity\x18\x04 \x01(\x02\x12\x14\n\x0cstroke_width\x18\x05 \x01(\x02"(\n\x05Point\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02"\xb4\x02\n\x0bMobjectData\x12\n\n\x02id\x18\x01 \x01(\t\x12!\n\x05style\x18\x02 \x01(\x0b\x32\x12.frameserver.Style\x12\x32\n\x04type\x18\x03 \x01(\x0e\x32$.frameserver.MobjectData.MobjectType\x12:\n\x17vectorized_mobject_data\x18\x04 \x01(\x0b\x32\x19.frameserver.VMobjectData\x12\x39\n\x12image_mobject_data\x18\x05 \x01(\x0b\x32\x1d.frameserver.ImageMobjectData\x12\x1b\n\x13root_mobject_offset\x18\x06 \x03(\x02".\n\x0bMobjectType\x12\x0c\n\x08VMOBJECT\x10\x00\x12\x11\n\rIMAGE_MOBJECT\x10\x01"H\n\x0cVMobjectData\x12"\n\x06points\x18\x01 \x03(\x0b\x32\x12.frameserver.Point\x12\x14\n\x0cneeds_redraw\x18\x02 \x01(\x08"c\n\x10ImageMobjectData\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x0e\n\x06height\x18\x02 \x01(\x02\x12\r\n\x05width\x18\x03 \x01(\x02\x12"\n\x06\x63\x65nter\x18\x04 \x01(\x0b\x32\x12.frameserver.Point"l\n\tFrameData\x12\x0e\n\x06remove\x18\x01 \x03(\t\x12%\n\x03\x61\x64\x64\x18\x02 \x03(\x0b\x32\x18.frameserver.MobjectData\x12(\n\x06update\x18\x03 \x03(\x0b\x32\x18.frameserver.MobjectData"\xfa\x01\n\rFrameResponse\x12*\n\nframe_data\x18\x01 \x01(\x0b\x32\x16.frameserver.FrameData\x12\x16\n\x0escene_finished\x18\x02 \x01(\x08\x12*\n\nanimations\x18\x03 \x03(\x0b\x32\x16.frameserver.Animation\x12&\n\x08updaters\x18\x04 \x03(\x0b\x32\x14.frameserver.Updater\x12\x17\n\x0f\x61nimation_index\x18\x05 \x01(\x05\x12\x18\n\x10\x61nimation_offset\x18\x06 \x01(\x02\x12\x1e\n\x16\x61ll_animations_tweened\x18\x07 \x01(\x08"\x0e\n\x0c\x45mptyRequest"\x0f\n\rEmptyResponse2\xf0\x01\n\x0b\x46rameServer\x12G\n\x0eGetFrameAtTime\x12\x19.frameserver.FrameRequest\x1a\x1a.frameserver.FrameResponse\x12P\n\x0e\x46\x65tchSceneData\x12\x19.frameserver.EmptyRequest\x1a#.frameserver.FetchSceneDataResponse\x12\x46\n\rScriptUpdated\x12\x19.frameserver.EmptyRequest\x1a\x1a.frameserver.EmptyResponseb\x06proto3',
)


_MOBJECTDATA_MOBJECTTYPE = _descriptor.EnumDescriptor(
    name="MobjectType",
    full_name="frameserver.MobjectData.MobjectType",
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name="VMOBJECT",
            index=0,
            number=0,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="IMAGE_MOBJECT",
            index=1,
            number=1,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=1161,
    serialized_end=1207,
)
_sym_db.RegisterEnumDescriptor(_MOBJECTDATA_MOBJECTTYPE)


_FETCHSCENEDATARESPONSE = _descriptor.Descriptor(
    name="FetchSceneDataResponse",
    full_name="frameserver.FetchSceneDataResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="scene",
            full_name="frameserver.FetchSceneDataResponse.scene",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="path",
            full_name="frameserver.FetchSceneDataResponse.path",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=34,
    serialized_end=107,
)


_SCENE = _descriptor.Descriptor(
    name="Scene",
    full_name="frameserver.Scene",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="name",
            full_name="frameserver.Scene.name",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="animations",
            full_name="frameserver.Scene.animations",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="background_color",
            full_name="frameserver.Scene.background_color",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=109,
    serialized_end=200,
)


_ANIMATION_ATTRIBUTETWEENDATA = _descriptor.Descriptor(
    name="AttributeTweenData",
    full_name="frameserver.Animation.AttributeTweenData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="attribute",
            full_name="frameserver.Animation.AttributeTweenData.attribute",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="start_data",
            full_name="frameserver.Animation.AttributeTweenData.start_data",
            index=1,
            number=2,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="end_data",
            full_name="frameserver.Animation.AttributeTweenData.end_data",
            index=2,
            number=3,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=446,
    serialized_end=523,
)

_ANIMATION_MOBJECTTWEENDATA = _descriptor.Descriptor(
    name="MobjectTweenData",
    full_name="frameserver.Animation.MobjectTweenData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="id",
            full_name="frameserver.Animation.MobjectTweenData.id",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="root_mobject_offset",
            full_name="frameserver.Animation.MobjectTweenData.root_mobject_offset",
            index=1,
            number=2,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=525,
    serialized_end=584,
)

_ANIMATION = _descriptor.Descriptor(
    name="Animation",
    full_name="frameserver.Animation",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="name",
            full_name="frameserver.Animation.name",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="duration",
            full_name="frameserver.Animation.duration",
            index=1,
            number=2,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="easing_function",
            full_name="frameserver.Animation.easing_function",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="attribute_tween_data",
            full_name="frameserver.Animation.attribute_tween_data",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="mobject_tween_data",
            full_name="frameserver.Animation.mobject_tween_data",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="flickered_mobject_ids",
            full_name="frameserver.Animation.flickered_mobject_ids",
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[
        _ANIMATION_ATTRIBUTETWEENDATA,
        _ANIMATION_MOBJECTTWEENDATA,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=203,
    serialized_end=584,
)


_UPDATER = _descriptor.Descriptor(
    name="Updater",
    full_name="frameserver.Updater",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="flickered_mobject_ids",
            full_name="frameserver.Updater.flickered_mobject_ids",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=586,
    serialized_end=626,
)


_FRAMEREQUEST = _descriptor.Descriptor(
    name="FrameRequest",
    full_name="frameserver.FrameRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="end_index",
            full_name="frameserver.FrameRequest.end_index",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="first_request",
            full_name="frameserver.FrameRequest.first_request",
            index=1,
            number=2,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="animation_index",
            full_name="frameserver.FrameRequest.animation_index",
            index=2,
            number=3,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="animation_offset",
            full_name="frameserver.FrameRequest.animation_offset",
            index=3,
            number=4,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=628,
    serialized_end=735,
)


_STYLE = _descriptor.Descriptor(
    name="Style",
    full_name="frameserver.Style",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="fill_color",
            full_name="frameserver.Style.fill_color",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="fill_opacity",
            full_name="frameserver.Style.fill_opacity",
            index=1,
            number=2,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="stroke_color",
            full_name="frameserver.Style.stroke_color",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="stroke_opacity",
            full_name="frameserver.Style.stroke_opacity",
            index=3,
            number=4,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="stroke_width",
            full_name="frameserver.Style.stroke_width",
            index=4,
            number=5,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=737,
    serialized_end=854,
)


_POINT = _descriptor.Descriptor(
    name="Point",
    full_name="frameserver.Point",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="x",
            full_name="frameserver.Point.x",
            index=0,
            number=1,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="y",
            full_name="frameserver.Point.y",
            index=1,
            number=2,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="z",
            full_name="frameserver.Point.z",
            index=2,
            number=3,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=856,
    serialized_end=896,
)


_MOBJECTDATA = _descriptor.Descriptor(
    name="MobjectData",
    full_name="frameserver.MobjectData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="id",
            full_name="frameserver.MobjectData.id",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="style",
            full_name="frameserver.MobjectData.style",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="type",
            full_name="frameserver.MobjectData.type",
            index=2,
            number=3,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="vectorized_mobject_data",
            full_name="frameserver.MobjectData.vectorized_mobject_data",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="image_mobject_data",
            full_name="frameserver.MobjectData.image_mobject_data",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="root_mobject_offset",
            full_name="frameserver.MobjectData.root_mobject_offset",
            index=5,
            number=6,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[
        _MOBJECTDATA_MOBJECTTYPE,
    ],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=899,
    serialized_end=1207,
)


_VMOBJECTDATA = _descriptor.Descriptor(
    name="VMobjectData",
    full_name="frameserver.VMobjectData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="points",
            full_name="frameserver.VMobjectData.points",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="needs_redraw",
            full_name="frameserver.VMobjectData.needs_redraw",
            index=1,
            number=2,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1209,
    serialized_end=1281,
)


_IMAGEMOBJECTDATA = _descriptor.Descriptor(
    name="ImageMobjectData",
    full_name="frameserver.ImageMobjectData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="path",
            full_name="frameserver.ImageMobjectData.path",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="height",
            full_name="frameserver.ImageMobjectData.height",
            index=1,
            number=2,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="width",
            full_name="frameserver.ImageMobjectData.width",
            index=2,
            number=3,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="center",
            full_name="frameserver.ImageMobjectData.center",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1283,
    serialized_end=1382,
)


_FRAMEDATA = _descriptor.Descriptor(
    name="FrameData",
    full_name="frameserver.FrameData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="remove",
            full_name="frameserver.FrameData.remove",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="add",
            full_name="frameserver.FrameData.add",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="update",
            full_name="frameserver.FrameData.update",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1384,
    serialized_end=1492,
)


_FRAMERESPONSE = _descriptor.Descriptor(
    name="FrameResponse",
    full_name="frameserver.FrameResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="frame_data",
            full_name="frameserver.FrameResponse.frame_data",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="scene_finished",
            full_name="frameserver.FrameResponse.scene_finished",
            index=1,
            number=2,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="animations",
            full_name="frameserver.FrameResponse.animations",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="updaters",
            full_name="frameserver.FrameResponse.updaters",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="animation_index",
            full_name="frameserver.FrameResponse.animation_index",
            index=4,
            number=5,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="animation_offset",
            full_name="frameserver.FrameResponse.animation_offset",
            index=5,
            number=6,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="all_animations_tweened",
            full_name="frameserver.FrameResponse.all_animations_tweened",
            index=6,
            number=7,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1495,
    serialized_end=1745,
)


_EMPTYREQUEST = _descriptor.Descriptor(
    name="EmptyRequest",
    full_name="frameserver.EmptyRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1747,
    serialized_end=1761,
)


_EMPTYRESPONSE = _descriptor.Descriptor(
    name="EmptyResponse",
    full_name="frameserver.EmptyResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1763,
    serialized_end=1778,
)

_FETCHSCENEDATARESPONSE.fields_by_name["scene"].message_type = _SCENE
_SCENE.fields_by_name["animations"].message_type = _ANIMATION
_ANIMATION_ATTRIBUTETWEENDATA.containing_type = _ANIMATION
_ANIMATION_MOBJECTTWEENDATA.containing_type = _ANIMATION
_ANIMATION.fields_by_name[
    "attribute_tween_data"
].message_type = _ANIMATION_ATTRIBUTETWEENDATA
_ANIMATION.fields_by_name[
    "mobject_tween_data"
].message_type = _ANIMATION_MOBJECTTWEENDATA
_MOBJECTDATA.fields_by_name["style"].message_type = _STYLE
_MOBJECTDATA.fields_by_name["type"].enum_type = _MOBJECTDATA_MOBJECTTYPE
_MOBJECTDATA.fields_by_name["vectorized_mobject_data"].message_type = _VMOBJECTDATA
_MOBJECTDATA.fields_by_name["image_mobject_data"].message_type = _IMAGEMOBJECTDATA
_MOBJECTDATA_MOBJECTTYPE.containing_type = _MOBJECTDATA
_VMOBJECTDATA.fields_by_name["points"].message_type = _POINT
_IMAGEMOBJECTDATA.fields_by_name["center"].message_type = _POINT
_FRAMEDATA.fields_by_name["add"].message_type = _MOBJECTDATA
_FRAMEDATA.fields_by_name["update"].message_type = _MOBJECTDATA
_FRAMERESPONSE.fields_by_name["frame_data"].message_type = _FRAMEDATA
_FRAMERESPONSE.fields_by_name["animations"].message_type = _ANIMATION
_FRAMERESPONSE.fields_by_name["updaters"].message_type = _UPDATER
DESCRIPTOR.message_types_by_name["FetchSceneDataResponse"] = _FETCHSCENEDATARESPONSE
DESCRIPTOR.message_types_by_name["Scene"] = _SCENE
DESCRIPTOR.message_types_by_name["Animation"] = _ANIMATION
DESCRIPTOR.message_types_by_name["Updater"] = _UPDATER
DESCRIPTOR.message_types_by_name["FrameRequest"] = _FRAMEREQUEST
DESCRIPTOR.message_types_by_name["Style"] = _STYLE
DESCRIPTOR.message_types_by_name["Point"] = _POINT
DESCRIPTOR.message_types_by_name["MobjectData"] = _MOBJECTDATA
DESCRIPTOR.message_types_by_name["VMobjectData"] = _VMOBJECTDATA
DESCRIPTOR.message_types_by_name["ImageMobjectData"] = _IMAGEMOBJECTDATA
DESCRIPTOR.message_types_by_name["FrameData"] = _FRAMEDATA
DESCRIPTOR.message_types_by_name["FrameResponse"] = _FRAMERESPONSE
DESCRIPTOR.message_types_by_name["EmptyRequest"] = _EMPTYREQUEST
DESCRIPTOR.message_types_by_name["EmptyResponse"] = _EMPTYRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FetchSceneDataResponse = _reflection.GeneratedProtocolMessageType(
    "FetchSceneDataResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _FETCHSCENEDATARESPONSE,
        "__module__": "frameserver_pb2",
        # @@protoc_insertion_point(class_scope:frameserver.FetchSceneDataResponse)
    },
)
_sym_db.RegisterMessage(FetchSceneDataResponse)

Scene = _reflection.GeneratedProtocolMessageType(
    "Scene",
    (_message.Message,),
    {
        "DESCRIPTOR": _SCENE,
        "__module__": "frameserver_pb2",
        # @@protoc_insertion_point(class_scope:frameserver.Scene)
    },
)
_sym_db.RegisterMessage(Scene)

Animation = _reflection.GeneratedProtocolMessageType(
    "Animation",
    (_message.Message,),
    {
        "AttributeTweenData": _reflection.GeneratedProtocolMessageType(
            "AttributeTweenData",
            (_message.Message,),
            {
                "DESCRIPTOR": _ANIMATION_ATTRIBUTETWEENDATA,
                "__module__": "frameserver_pb2",
                # @@protoc_insertion_point(class_scope:frameserver.Animation.AttributeTweenData)
            },
        ),
        "MobjectTweenData": _reflection.GeneratedProtocolMessageType(
            "MobjectTweenData",
            (_message.Message,),
            {
                "DESCRIPTOR": _ANIMATION_MOBJECTTWEENDATA,
                "__module__": "frameserver_pb2",
                # @@protoc_insertion_point(class_scope:frameserver.Animation.MobjectTweenData)
            },
        ),
        "DESCRIPTOR": _ANIMATION,
        "__module__": "frameserver_pb2",
        # @@protoc_insertion_point(class_scope:frameserver.Animation)
    },
)
_sym_db.RegisterMessage(Animation)
_sym_db.RegisterMessage(Animation.AttributeTweenData)
_sym_db.RegisterMessage(Animation.MobjectTweenData)

Updater = _reflection.GeneratedProtocolMessageType(
    "Updater",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATER,
        "__module__": "frameserver_pb2",
        # @@protoc_insertion_point(class_scope:frameserver.Updater)
    },
)
_sym_db.RegisterMessage(Updater)

FrameRequest = _reflection.GeneratedProtocolMessageType(
    "FrameRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _FRAMEREQUEST,
        "__module__": "frameserver_pb2",
        # @@protoc_insertion_point(class_scope:frameserver.FrameRequest)
    },
)
_sym_db.RegisterMessage(FrameRequest)

Style = _reflection.GeneratedProtocolMessageType(
    "Style",
    (_message.Message,),
    {
        "DESCRIPTOR": _STYLE,
        "__module__": "frameserver_pb2",
        # @@protoc_insertion_point(class_scope:frameserver.Style)
    },
)
_sym_db.RegisterMessage(Style)

Point = _reflection.GeneratedProtocolMessageType(
    "Point",
    (_message.Message,),
    {
        "DESCRIPTOR": _POINT,
        "__module__": "frameserver_pb2",
        # @@protoc_insertion_point(class_scope:frameserver.Point)
    },
)
_sym_db.RegisterMessage(Point)

MobjectData = _reflection.GeneratedProtocolMessageType(
    "MobjectData",
    (_message.Message,),
    {
        "DESCRIPTOR": _MOBJECTDATA,
        "__module__": "frameserver_pb2",
        # @@protoc_insertion_point(class_scope:frameserver.MobjectData)
    },
)
_sym_db.RegisterMessage(MobjectData)

VMobjectData = _reflection.GeneratedProtocolMessageType(
    "VMobjectData",
    (_message.Message,),
    {
        "DESCRIPTOR": _VMOBJECTDATA,
        "__module__": "frameserver_pb2",
        # @@protoc_insertion_point(class_scope:frameserver.VMobjectData)
    },
)
_sym_db.RegisterMessage(VMobjectData)

ImageMobjectData = _reflection.GeneratedProtocolMessageType(
    "ImageMobjectData",
    (_message.Message,),
    {
        "DESCRIPTOR": _IMAGEMOBJECTDATA,
        "__module__": "frameserver_pb2",
        # @@protoc_insertion_point(class_scope:frameserver.ImageMobjectData)
    },
)
_sym_db.RegisterMessage(ImageMobjectData)

FrameData = _reflection.GeneratedProtocolMessageType(
    "FrameData",
    (_message.Message,),
    {
        "DESCRIPTOR": _FRAMEDATA,
        "__module__": "frameserver_pb2",
        # @@protoc_insertion_point(class_scope:frameserver.FrameData)
    },
)
_sym_db.RegisterMessage(FrameData)

FrameResponse = _reflection.GeneratedProtocolMessageType(
    "FrameResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _FRAMERESPONSE,
        "__module__": "frameserver_pb2",
        # @@protoc_insertion_point(class_scope:frameserver.FrameResponse)
    },
)
_sym_db.RegisterMessage(FrameResponse)

EmptyRequest = _reflection.GeneratedProtocolMessageType(
    "EmptyRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _EMPTYREQUEST,
        "__module__": "frameserver_pb2",
        # @@protoc_insertion_point(class_scope:frameserver.EmptyRequest)
    },
)
_sym_db.RegisterMessage(EmptyRequest)

EmptyResponse = _reflection.GeneratedProtocolMessageType(
    "EmptyResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _EMPTYRESPONSE,
        "__module__": "frameserver_pb2",
        # @@protoc_insertion_point(class_scope:frameserver.EmptyResponse)
    },
)
_sym_db.RegisterMessage(EmptyResponse)


_FRAMESERVER = _descriptor.ServiceDescriptor(
    name="FrameServer",
    full_name="frameserver.FrameServer",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=1781,
    serialized_end=2021,
    methods=[
        _descriptor.MethodDescriptor(
            name="GetFrameAtTime",
            full_name="frameserver.FrameServer.GetFrameAtTime",
            index=0,
            containing_service=None,
            input_type=_FRAMEREQUEST,
            output_type=_FRAMERESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="FetchSceneData",
            full_name="frameserver.FrameServer.FetchSceneData",
            index=1,
            containing_service=None,
            input_type=_EMPTYREQUEST,
            output_type=_FETCHSCENEDATARESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="ScriptUpdated",
            full_name="frameserver.FrameServer.ScriptUpdated",
            index=2,
            containing_service=None,
            input_type=_EMPTYREQUEST,
            output_type=_EMPTYRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_FRAMESERVER)

DESCRIPTOR.services_by_name["FrameServer"] = _FRAMESERVER

# @@protoc_insertion_point(module_scope)
